#!/usr/bin/perl -w

#############################################################################
#                                                                           #
# File:  finddups                                                           #
#                                                                           #
# Written by Petr Hekerle 1999                                              #
#                                                                           #
#############################################################################

use POSIX;
use strict;
use DBI;
use MD5;


my $md5 = new MD5;
my @stat;
my @row;
my $st_dev = -1;
my $new;
my $debug;
my $dontmake;
my $volume;
my $changedbi;
my $date = time;
my $dbh;
my $select;
my $selectcrc;
my $insert;
my $update;
my $updatedate;
my $dbname = "finddupsdb";
my $user = "finddups-user";
my $password = "PaSSWord";

#############################################################################
# Help                                                                      #
#############################################################################

sub Help
 {
 printf("Usage:\tfinddups [-dxvf] [-u dbname user password] directory1 [directory2 ...]\n");
 printf("\tDuplicite files in the given trees substitutes by a hard-link.\n\n");
 printf("\tParameters:\n");
 printf("\t-d\tPrints the duplicites and makes the hardlink\n");
 printf("\t-x\tPrints the duplicites and does not make the hardlink\n");
 printf("\t-v\tDo not cross the volume boundary\n");
 printf("\t-f\tDeletes old data in the database.\n");
 printf("\t-u\tName of the database, username and password for the DB engine\n");
 exit;
 }

#############################################################################
# dbiconect                                                                 #
#############################################################################

sub dbiconect
 {
 $dbh = DBI->connect("dbi:mysql:$dbname", $user, $password) || die $DBI::errstr;
 $select = $dbh->prepare("select * from finddups where name=?");
 $selectcrc = $dbh->prepare("select * from finddups where md5sum=? and sumdate=?");
 $insert = $dbh->prepare("insert into finddups values(?, ?, ?, ?, ?, ?, ?)");
 $update = $dbh->prepare("update finddups set size=?, date=?, md5sum=?, sumdate=?, inode=?
			    where name=?");
 $updatedate = $dbh->prepare("update finddups set sumdate=?, inode=? where name=?");
 }

#############################################################################
# getdir(root);                                                             #
#############################################################################

sub getdir
 {
 my ($rootdir) = @_;

#if ($debug || $dontmake) { printf("$rootdir\n"); }

 if ( !opendir(DIR, $rootdir) )
     {
     printf(STDERR "Adresar $rootdir nebude prohlizen - nelze ho otevrit : $!\n");
     return;
     }
 else{
     my $st_detect;
     foreach (sort readdir(DIR))
	 {
	 next if (/^\.\.?$/); # preskocit pokud je "." nebo ".."
	 my $filename = $_;
  	 $filename = "$rootdir/$filename";

   	 $filename =~ s/\/\//\//;

	 next if (-l $filename);
	 @stat = stat "$filename";
	 next if ($stat[7] == 0); # preskocit pokud je size = 0

	 if ($volume && !$st_detect)
	     {
	     ++$st_detect;
	     if ($st_dev == -1) { $st_dev = $stat[0]; } # priradi cislo svazku do $st_dev
	     if ($stat[0] != $st_dev)
		 {
		 printf(STDERR "Adresar $rootdir nebude prohlizen - je v jinem svazku\n");
		 return;
		 }
	     }
   	 if (-f "$filename" && !-l "$filename")
	     {
	     my $filecrc;
	     
	     if ( open(FILE,"$filename") )
		 {
		 my $makeCRC = 1;
		 $select->execute($filename);
		 while ( @row = $select->fetchrow_array )
		     {
		     $makeCRC = 2;
		     if ($row[1] == $stat[7] && $row[2] == $stat[9])
			 {
			 $filecrc = $row[3];
			 $makeCRC = 0;
			 }
		     }

		 if ($makeCRC != 0)
		     {
		     $md5->reset;
		     $md5->addfile(\*FILE);
		     $filecrc = $md5->hexdigest();

             	     $filecrc =~ s/\n$//; # odstrani pripadne konce radku
		     }

		 my $inode = $stat[1];
		 my $next = 0;
		 $selectcrc->execute($filecrc, $date);
		 while ( @row = $selectcrc->fetchrow_array )
		     {
		     if ($row[0] ne $filename && $row[1] == $stat[7] &&
			 $row[5] == $stat[0] && $row[6] != $stat[1])
			 {
			 my $filecrcdups;
			 my $nodups = 0;
			 if (!$makeCRC && !$next) #jestlize nebyla zjistena md5
			     {
			     ++$next;
			     $md5->reset;
			     $md5->addfile(\*FILE);
			     $filecrcdups = $md5->hexdigest();
			     $filecrcdups =~ s/\n$//;
			     if ($filecrc ne $filecrcdups)
				 {
				 printf(STDERR "Abnormalni zmena md5sum souboru: $filename\n");
				 printf(STDERR " stara: $filecrc nova: $filecrcdups\n");
				 $filecrc = $filecrcdups;
				 $makeCRC = 2;
				 ++$nodups;
				 }
			     }
			 if ( open(DUPS, "$row[0]") )
			     {
			     $md5->reset;
			     $md5->addfile(\*DUPS);
			     $filecrcdups = $md5->hexdigest();
			     $filecrcdups =~ s/\n$//;
			     if ($row[3] ne $filecrcdups)
				 {
				 printf(STDERR "Abnormalni zmena md5sum souboru: $row[0]\n");
				 printf(STDERR " stara: $row[3] nova: $filecrcdups\n");
				 if ($filecrc ne $filecrcdups)
				     { --$nodups }
				 $update->execute($row[1], $row[2], $filecrcdups, $date,
						  $row[6], $row[0]);
				 }
			     close(DUPS);
			     }
			 else{ printf(STDERR "Soubor $row[0] nelze otevrit : $!\n"); }
			 unless ($nodups)
			     {
			     if ($debug || $dontmake)
				 {
				 printf("Duplic: $filename\n");
				 printf("\t$row[0]\n");
				 }
			     unless ($dontmake)
				 {
				 # Provede prislusnou operaci s duplicitnimi soubory
				 #Lze jednoruse zmenit na symlink nabo na smazani souboru
				 unlink($filename);
				 link($row[0], $filename);
				 $inode = $row[6];
				 }
			     }
			 }
		     }
		 if ($makeCRC == 0)
		     { $updatedate->execute($date, $inode, $filename); }
		 if ($makeCRC == 1)
		     { $insert->execute($filename, $stat[7], $stat[9], $filecrc, $date,
					$stat[0], $inode); }
		 if ($makeCRC == 2)
		     { $update->execute($stat[7], $stat[9], $filecrc, $date, $inode,
					$filename); }

		 close(FILE);
		 }
	     else{ printf(STDERR "Soubor $filename nelze otevrit : $!\n"); }
	     }
	 if (-d "$filename") { &getdir("$filename"); }
	 }
     close(DIR);
     }
 }

#############################################################################
# Main routine begins.                                                      #
#############################################################################

if($#ARGV==-1) { &Help; }
if(($#ARGV>=0) && ($ARGV[0] =~ /^-/))
    {
    foreach my $arg (@ARGV)
	{
	if ($arg =~ /^-/)
            {
            if ($arg =~ /f/) { $new = 1; }
            if ($arg =~ /d/) { $debug = 1; }
            if ($arg =~ /x/) { $dontmake = 1; }
            if ($arg =~ /v/) { $volume = 1; }
	    if ($arg =~ /u/) { $changedbi = 1; }
            if ($arg =~ /\?/){ &Help; }
            if ($arg =~ /\h/){ &Help; }
            }
	}
    shift;
    }

if ($changedbi)
    {
    if($#ARGV==-1)
	{
	printf("Zadej parametr: dbname user password\n");
	printf("Usage:\tfinddups -u dbname user password directory1 [directory2 ...]\n");
	exit;
	}
    $dbname = $ARGV[0];
    shift;
    if($#ARGV==-1)
	{
	printf("Zadej parametr: user password\n");
	printf("Usage:\tfinddups -u dbname user password directory1 [directory2 ...]\n");
	exit;
	}
    $user = $ARGV[0];
    shift;
    if($#ARGV==-1)
	{
	printf("Zadej parametr: password\n");
	printf("Usage:\tfinddups -u dbname user password directory1 [directory2 ...]\n");
	exit;
	}
    $password = $ARGV[0];
    shift;
    }

# Pokud nebyl zadan zadny adresar, nastav aktualni adresar
if($#ARGV==-1) { @ARGV = `pwd`; }
&dbiconect();

if ($new)
    {
    my $delete = $dbh->do("delete from finddups where sumdate!=$date");
    printf("Obsah databaze byl vymazan ($delete polozek)\n");
    }

foreach my $dir (@ARGV)
    {
    $dir =~ s/^\.\///;
    $dir =~ s/^\.$//;
    unless ($dir =~ /^\//)
	{
	if ($dir ne "") { $dir = `pwd`.'/'.$dir }
	else { $dir = `pwd` }
	}
    $dir =~ s/\n//;

    &getdir($dir);
    }

$dbh->disconnect  || warn $dbh->errstr;
